from datetime import datetime

import mesa
from mesa.datacollection import DataCollector 

from src.Agent import Agent
from src.Navigation import Navigation
from src.helper import *


class Model(mesa.Model):
    """A model with some number of agents."""

    def __init__(self, number_of_agents, weight_terrain_map, weight_base_cost, weight_record_cost, safe_to_disk):
        print("Starting Simulation")

        # run_name
        self.run_name = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")

        # if true the records will be saved to disk
        self.safe_to_disk = safe_to_disk

        # parameter
        self.weight_terrain_map = weight_terrain_map
        self.weight_base_cost = weight_base_cost

        super().__init__()
        self.num_agents = number_of_agents
        # Create scheduler and assign it to the model
        self.schedule = mesa.time.RandomActivation(self)
        self.nav = Navigation(self, weight_terrain_map, weight_base_cost, weight_record_cost)
        self.grid = mesa.space.MultiGrid(self.nav.width, self.nav.height, False)

        #Datacollector moving agents
        self.datacollector_moving = DataCollector(
            {
                "Moving Agents": Model.current_moving_agents
                #"Non Healthy Agents": FightingModel.current_non_healthy_agents,
            }
        )

        #Datacollector agent tasks
        self.datacollector_task = DataCollector(
            {
                "Home": Model.task_home,
                "Work": Model.task_work,
                "Church": Model.task_church,
                "Market": Model.task_market,
                "Random": Model.task_random
            }
        )
        
        # Create agents
        for i in range(self.num_agents):
            a = Agent(i, self)
            # Add the agent to the scheduler
            self.schedule.add(a)

            pos_on_street = self.nav.random_street_pixel()

        print("Simulation started")

    # returns the current step count
    def get_step_count(self):
        return self.schedule.steps

    def step(self):


        """Advance the model by one step."""
        # print("Current day: ", tick_to_day(self.get_step_count()), " | Hour: ", tick_to_hour(self.get_step_count()))

        if self.get_step_count() != 0 and self.get_step_count() % (TICKS_PER_HOUR * 24) == 0:
            self.nav.finish_day(self.safe_to_disk)

        # let agents make one step
        self.schedule.step()

        #collect data
        self.datacollector_moving.collect(self)
        self.datacollector_task.collect(self)

    #methods for Datacollector
    
    @staticmethod
    def current_moving_agents(model) -> int:
        """Return the total number of currently moving agents

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of moving agents
        """
        return sum([1 for agent in model.schedule.agents if agent.isMoving == True])

    @staticmethod
    def task_home(model) -> int:
        """
        Return the total number of agents going home

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of agents going home
        """
        return sum([1 for agent in model.schedule.agents if agent.behavior.get_task() == "home"])
    
    @staticmethod
    def task_work(model) -> int:
        """
        Return the total number of agents going to work

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of agents going to work
        """
        return sum([1 for agent in model.schedule.agents if agent.behavior.get_task() == "work"]) + Model.task_home(model)

    @staticmethod
    def task_random(model) -> int:
        """
        Return the total number of agents with random target

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of agents with random target
        """
        return sum([1 for agent in model.schedule.agents if agent.behavior.get_task() == "random_target"]) + Model.task_work(model)

    
    @staticmethod
    def task_church(model) -> int:
        """
        Return the total number of agents going to church

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of agents going to work
        """
        return sum([1 for agent in model.schedule.agents if agent.behavior.get_task() == "church"]) + Model.task_random(model)
        
    @staticmethod
    def task_market(model) -> int:
        """
        Return the total number of agents going to the market

        Args:
            model (Model): The simulation model

        Returns:
            int: Number of agents going to the market
        """
        return sum([1 for agent in model.schedule.agents if agent.behavior.get_task() == "market"]) + Model.task_church(model)

   
