import numpy as np
import tifffile

from config import *



class MovementRecorder:
    def __init__(self, width, height):
        self.width = width
        self.height = height

        # dictionary of all records
        self.records = dict()
        # dictionary of all normalized records (0-1)
        self.records_normalized = dict()
        # dictionary of all normalized averaged other time
        self.records_time_average = dict()

    # creates a new record and a corresponding normalized record
    def add_record(self, name):
        # check
        if self.records.__contains__(name):
            Exception("Record name must be unique")
        if len(name) == 0:
            Exception("Name can't be empty")

        # create new records
        self.records[name] = np.zeros(shape=(self.width, self.height))
        self.records_normalized[name] = np.zeros(shape=(self.width, self.height))
        self.records_time_average[name] = np.zeros(shape=(self.width, self.height))

    # add the given amount at the position in the given record
    def increase_val_in_record(self, record_name, x, y, amount):

        if amount < 0:
            Exception("amount must be positive")

        # check
        if not self.records.__contains__(record_name):
            Exception("Unknown record: ", record_name)

        self.records[record_name][x][y] += amount

    # clears the record with the given name
    def clear_record(self, record_name):
        self.records[record_name] = np.zeros(shape=(self.width, self.height))

    # clears the normalized record with the given name
    def clear_record_normalized(self, record_name):
        self.records[record_name] = np.zeros(shape=(self.width, self.height))

    # divides all record data by the max value and writs it to the record_normalized
    # overwrites the previous data
    def normalize_record(self, record_name):
        # print("Normalize:", record_name)

        selected_record = self.records[record_name]
        selected_record_normalized = self.records_normalized[record_name]

        max_val = np.max(selected_record)

        # copy and divide the values by max if max is not 0
        if max_val > 0:
            for i in range(self.width):
                for j in range(self.height):
                    val = selected_record[i][j]
                    val_normalized = val / max_val
                    selected_record_normalized[i][j] = val_normalized
        else:
            # record was all 0 -> just clear normalized record
            self.clear_record_normalized(record_name)

        # clear record
        self.clear_record(record_name)

    # returns the record with the given name
    def get_record(self, record_name):
        return self.records[record_name]

    # returns the normalized record with the given name
    def get_record_normalized(self, record_name):
        return self.records_normalized[record_name]

    # return the value from the record at xy
    def get_sample_time_average(self, record_name, x, y):
        return self.records_time_average[record_name][x][y]

    def update_average(self, record_name):
        # print("Update:", record_name)

        selected_record_time_averaged = self.records_time_average[record_name]
        selected_record_normalized = self.records_normalized[record_name]

        max_val = np.max(selected_record_time_averaged)

        # first time (selected_record_time_averaged is all 0) the normalized will be just copied
        if max_val > 0:
            for i in range(self.width):
                for j in range(self.height):
                    val_n = selected_record_normalized[i][j]
                    val_t = selected_record_time_averaged[i][j]
                    val_new = val_t * (1 - TIME_AVERAGED_CHANGE_SPEED) + val_n * TIME_AVERAGED_CHANGE_SPEED
                    selected_record_time_averaged[i][j] = val_new
        else:
            self.records_time_average[record_name] = np.copy(selected_record_normalized)

    def save_record(self, record_name, record_type_name, location, file_name):
        record_types = {
            "short_term": self.records,
            "normalized": self.records_normalized,
            "time_averaged": self.records_time_average,
        }

        if not record_types.__contains__(record_type_name):
            Exception("Unknown record type: ", record_type_name)

        records = record_types[record_type_name]

        if not records.__contains__(record_name):
            Exception("Unknown record: ", record_name)

        selected_record = records[record_name]

        if not isinstance(selected_record, np.ndarray) or selected_record.shape != (256, 256):
            raise ValueError("Selected record is not a 2D numpy array with 256x256 resolution.")

        # Construct the full file path
        full_path = f"{location}/{file_name+'.tiff'}"

        # Rotate the image 90 degrees counterclockwise
        selected_record_rotated = np.rot90(selected_record)

        # Scale the values to 0-255 range for 8-bit TIFF
        selected_record_scaled = (selected_record_rotated * 255).astype(np.uint8)

        # Save the numpy array as a tiff file
        tifffile.imwrite(full_path, selected_record_scaled)

        # print(f"Record {record_name} of type {record_type_name} has been saved to {full_path}")

