import os

import networkx as nx
import numpy as np
from PIL import Image

from src.MovementRecorder import MovementRecorder
from src.helper import *


class Navigation:
    def __init__(self, model, weight_terrain_map, weight_base_cost, weight_record_cost):

        # parameter
        self.weight_terrain_map = weight_terrain_map
        self.weight_base_cost = weight_base_cost
        self.weight_record_cost = weight_record_cost

        # array of pixel on streets (x, y)
        self.street_pixels = []
        # outgoing streets
        self.outgoing_connections = [
            (36, 32),
            (21, 83),
            (43, 104),
            (0, 192),
            (36, 255),
            (101, 255),
            (156, 254),
            (255, 188),
            (238, 167),
            (233, 98)
        ]
        # coordinates of the two churches in the area
        self.churches = [(182, 133), (33, 80)]
        # mid-point of the market
        self.market_location = (105, 120)
        # street pixel around the market
        self.market_pixels = []

        # 2d arrays which contains the streets and the cost map
        self.streets_np_array = None
        self.terrain_np_array = None
        # networkx graph of the streets
        self.G = None

        # size
        self.width = 0
        self.height = 0

        # link to the model
        self.model = model

        # load data
        self.load_images()

        # create graph
        self.create_street_graph()

        # calculate market pixel
        self.calc_market_pixel()

        # recorder for the movement of the agent. stores the traffic of each pixel
        self.move_rec = MovementRecorder(self.width, self.height)

        self.move_rec.add_record("day")
        self.move_rec.add_record("night")

        self.path_costs_current_day = []
        self.path_length_current_day = []

    # load the cost map and street
    def load_images(self):
        image_terrain = Image.open('data/street_binary_cost_256.tif')
        image_streets = Image.open('data/street_binary_256.tif')

        # safe dimension
        self.width = image_streets.width
        self.height = image_streets.height
        print("Map | Size: ", self.width, self.height)

        # 2d arrays which contains the streets and the cost map
        self.terrain_np_array = np.zeros((self.width, self.height))
        self.streets_np_array = np.zeros((self.width, self.height))

        # close hole in the street array
        image_streets.putpixel((98, 255), image_streets.getpixel((97, 255)))
        image_terrain.putpixel((98, 255), image_terrain.getpixel((97, 255)))

        # load the data
        for x in range(self.width):
            for y in range(self.height):

                # ------------- fill streets array -------------
                street_val = image_streets.getpixel((x, self.height - 1 - y))  # switch origin from top to bottom
                # handle self create tiff files
                if type(street_val) is tuple:
                    street_val = street_val[0] / 255
                if street_val == 1:
                    # is street
                    self.streets_np_array[x, y] = 1
                    self.street_pixels.append((x, y))
                else:
                    # is not street
                    self.streets_np_array[x, y] = 0

                # ------------- fill cost map -------------
                cost_val = image_terrain.getpixel((x, self.height - 1 - y))  # switch origin from top to bottom
                # handle self create tiff files
                if type(cost_val) is tuple:
                    cost_val = cost_val[0] / 255

                self.terrain_np_array[x, y] = cost_val

        print("Street Pixel Count:", len(self.street_pixels))
        print("Outgoing Connection Count:", len(self.outgoing_connections))

    # prints ascii street map to the console
    # hashtags = streets, underscore = not street
    def print_streets(self):
        for y in range(self.height - 1, -1, -1):
            line = ""
            for x in range(self.width):
                if self.streets_np_array[x, y] == 1:
                    line += "#"
                else:
                    line += "_"
            print(line)

    def print_terrain_ascii(self):
        # Define ASCII characters for different ranges
        ascii_chars = np.array(['.', ':', '|', '*', '#'])

        min_val = np.min(self.terrain_np_array)
        max_val = np.max(self.terrain_np_array)

        if min_val < 0:
            Exception("Min cost in terrain is negative")

        for y in range(self.height - 1, -1, -1):
            line = ""
            for x in range(self.width):
                val = self.terrain_np_array[x, y]
                val_normalized = val
                if max_val > 0:
                    val_normalized /= max_val
                char_index = int(math.floor(val_normalized * 4))
                line += ascii_chars[char_index]
            print(line)

    # returns random pixel coordinate on a street
    def random_street_pixel(self):
        return self.street_pixels[random.randint(0, len(self.street_pixels) - 1)]

    # returns random pixel coordinate of an outgoing connection
    def random_outgoing_connection_pixel(self):
        return self.outgoing_connections[random.randint(0, len(self.outgoing_connections) - 1)]

    # returns the nearest church to the given coordinate
    def nearest_church(self, home_coordinates):
        shortest_distance = float('inf')
        nearest_church_co = None

        for church_coords in self.churches:
            distance = math.dist(church_coords, home_coordinates)

            if distance < shortest_distance:
                shortest_distance = distance
                nearest_church_co = church_coords

        return nearest_church_co

    # returns random market pixel
    def random_market_pixel(self):
        return self.market_pixels[random.randint(0, len(self.market_pixels) - 1)]

    # heuristic method for a*
    # used to approximate the distance to the target
    def heuristic(self, a, b):
        # Euclidean distance as the heuristic
        return math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)

    # this function calculate the price for moving.
    # moving diagonal cost the root of 2 or sqrt(1^2 + 1^2)
    def weight_function(self, u, v, d):
        ux, uy = u
        vx, vy = v
        if abs(ux - vx) == 1 and abs(uy - vy) == 1:  # Diagonal move
            move_cost = math.sqrt(2)
        else:  # Horizontal or vertical move
            move_cost = 1

        terrain_cost = move_cost * self.terrain_np_array[vx, vy] * self.weight_terrain_map
        base_cost = move_cost * self.weight_base_cost
        record_cost = move_cost * self.move_rec.get_sample_time_average(self.current_record_name(), vx, vy) * self.weight_record_cost
        total_cost = terrain_cost + base_cost + record_cost

        return total_cost

    def create_street_graph(self):
        # Create a full 2D grid graph
        self.G = nx.grid_2d_graph(self.width, self.height, create_using=nx.Graph())

        # Remove nodes that are not street pixels
        for x in range(self.width):
            for y in range(self.height):
                if self.streets_np_array[x, y] != 1:
                    self.G.remove_node((x, y))

        # Add diagonal edges.
        # For simplicity edges will be added twice. NetworkX ignores duplicat edges.
        diagonal_offsets = [(-1, -1), (1, -1), (-1, 1), (1, 1)]

        for x in range(1, self.width - 1):
            for y in range(1, self.height - 1):
                # skip none street pixel
                if not self.streets_np_array[x, y] == 1:
                    continue

                for offset in diagonal_offsets:
                    # diagonal coords
                    dx = x + offset[0]
                    dy = y + offset[1]

                    # skip if the diagonal pixel is not a street pixel
                    if not self.streets_np_array[dx, dy] == 1:
                        continue

                    self.G.add_edge((x, y), (dx, dy))

    # Calculate the cheapest path between start_coords and end_coords using A* algorithm.
    def calc_path(self, start_coords, end_coords):
        # Calculate the shortest path using A* algorithm
        try:
            shortest_path = nx.astar_path(self.G, source=start_coords, target=end_coords, heuristic=self.heuristic,
                                          weight=self.weight_function)

            # calc cost for paths longer 2 (remove not moving agents)
            if len(shortest_path) > 2:
                cost = 0

                for i in range(0, len(shortest_path) - 2):
                    u = shortest_path[0]
                    v = shortest_path[i]
                    cost += self.weight_function(u, v, 0)

                self.path_costs_current_day.append(cost)
                self.path_length_current_day.append(len(shortest_path))

            return shortest_path
        except nx.NetworkXNoPath:
            print(f"No path found from {start_coords} to {end_coords}")
            return []

    def current_record_name(self):
        current_tick = self.model.get_step_count()
        current_hour = tick_to_hour(current_tick)
        # night is from 0-4 and 22-23 or 00:00-04:00 and 22:00-00:00
        return "night" if current_hour < 4 or current_hour > 21 else "day"

    def save_step_to_record(self, x, y, amount):

        # select/reselect record based on current tick
        self.move_rec.increase_val_in_record(self.current_record_name(), x, y, amount)

    # calculate all pixels surrounding the market
    def calc_market_pixel(self):
        unvisited_pixel = [self.market_location]
        visited_pixel = set()
        surrounding = set()

        while len(unvisited_pixel) > 0:
            # get first from list
            current_pixel = unvisited_pixel.pop()
            visited_pixel.add(current_pixel)

            offsets = [(-1, 0), (0, 1), (1, 0), (0, -1)]

            for offset in offsets:
                sample_x = current_pixel[0] + offset[0]
                sample_y = current_pixel[1] + offset[1]

                value = self.streets_np_array[sample_x][sample_y]

                if value == 1:
                    surrounding.add((sample_x, sample_y))
                elif not visited_pixel.__contains__((sample_x, sample_y)):
                    unvisited_pixel.append((sample_x, sample_y))
        self.market_pixels = list(surrounding)

    def create_run_folder_if_needed(self):
        # Get the current working directory
        current_working_directory = os.getcwd()

        # Create the path for the run folder
        run_folder_path = os.path.join(current_working_directory, "runs", self.model.run_name)

        # Check if the folder exists, and if not, create it
        if not os.path.exists(run_folder_path):
            os.makedirs(run_folder_path)


    # finish day by updating the recorder
    def finish_day(self, safe_to_disk):

        record_names = ["day", "night"]

        # path for saving run
        # create path if needed
        self.create_run_folder_if_needed()
        path = os.path.join(os.getcwd(), "runs", self.model.run_name)

        day_index = tick_to_day(self.model.get_step_count())

        for rec_name in record_names:
            self.move_rec.normalize_record(rec_name)

            filename = rec_name + "_n_" + str(day_index)
            if safe_to_disk:
                self.move_rec.save_record(rec_name, "normalized", path, filename)

            self.move_rec.update_average(rec_name)

            filename = rec_name + "_ta_" + str(day_index)

            if safe_to_disk:
                self.move_rec.save_record(rec_name, "time_averaged", path, filename)

        # calc average path cost
        average_path_cost = np.average(self.path_costs_current_day)
        # print("Average path cost is: ", average_path_cost)
        self.path_costs_current_day = []
        average_path_length = np.average(self.path_length_current_day)
        # print("Average path length is: ", average_path_length)
        self.path_length_current_day = []
        print(average_path_cost / average_path_length / self.weight_record_cost)